# coding=utf-8
# author=uliontse

"""
Copyright 2020 UlionTse

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from mlgb.tf.configs import (
    tf,
    MVPoolModeList,
    SampleModeList,
)
from mlgb.tf.components.linears import (
    TaskLayer,
)
from mlgb.tf.components.trms import (
    LabelAttentionLayer,
)
from mlgb.tf.components.retrieval import (
    SampledSoftmaxLossLayer,
)
from mlgb.tf.modules.matching import (
    DeepStructuredSemanticModelUserEmbeddingLayer,
    DeepStructuredSemanticModelItemEmbeddingLayer,
    NeuralCollaborativeFilteringLayer,
    MultiInterestNetworkWithDynamicRoutingLayer,
    UserInputs_MultiInterestNetworkWithDynamicRoutingLayer,
)
from mlgb.error import MLGBError


__all__ = [
    'NCF',
    'MatchFM', 'EBR', 'DSSM', 'YoutubeDNN', 'MIND',
]


class NCF(tf.keras.Model):
    def __init__(self, feature_names, task='multiclass:100', seed=None,
                 user_inputs_if_multivalued=False, user_inputs_if_sequential=False,
                 item_inputs_if_multivalued=False,
                 embed_dim=32, embed_l2=0.0, embed_initializer=None,
                 pool_mv_mode='Pooling:average', pool_mv_axis=2, pool_mv_l2=0.0, pool_mv_initializer=None,
                 pool_seq_mode='Pooling:average', pool_seq_axis=1, pool_seq_l2=0.0, pool_seq_initializer=None,
                 sample_mode='Sample:batch', sample_num=None, sample_item_distribution_list=None,
                 sample_fixed_unigram_frequency_list=None, sample_fixed_unigram_distortion=1.0,
                 model_embeds_if_l2_norm=True, model_result_temperature_ratio=None,
                 user_dnn_hidden_units=(), item_dnn_hidden_units=(),
                 dnn_hidden_units=(1024, 512), dnn_activation='tanh', dnn_dropout=0.0,
                 dnn_if_bn=False, dnn_if_ln=False, dnn_l2=0.0, dnn_initializer=None,
                 ):
        """
        Model Name: NCF(NeuralCollaborativeFiltering)
        Paper Team: NUS(Xiangnan He(NUS), etc)
        Paper Year: 2017
        Paper Name: <Neural Collaborative Filtering>
        Paper Link: https://arxiv.org/pdf/1708.05031.pdf

        Task Inputs Parameters:
            :param feature_names: tuple(tuple(dict)), must. Embedding need vocabulary size and custom embed_dim of features.
            :param task: str, default 'multiclass:100'. Union['binary', 'regression', 'multiclass:{int}']
            :param seed: Optional[int], default None.
            :param user_inputs_if_multivalued: bool, default False.
            :param user_inputs_if_sequential: bool, default False.
            :param item_inputs_if_multivalued: bool, default False.
            :param embed_dim: int, default 32.
            :param embed_l2: float, default 0.0.
            :param embed_initializer: Optional[str], default None. When None, activation judge first, he_uniform end.
            :param pool_mv_mode: str, default 'Pooling:average'. Pooling mode of multivalued inputs. Union[
                                'Attention', 'Weighted', 'Pooling:max', 'Pooling:average', 'Pooling:sum']
            :param pool_mv_axis: int, default 2. Pooling axis of multivalued inputs.
            :param pool_mv_l2: float, default 0.0. When pool_mv_mode is in ('Weighted', 'Attention'), it works.
            :param pool_mv_initializer: Optional[str], default None. When None, activation judge first,
                                he_uniform end. When pool_mv_mode is in ('Weighted', 'Attention'), it works.
            :param pool_seq_mode: str, default 'Pooling:average'. Pooling mode of sequential inputs. Union[
                                'Attention', 'Weighted', 'Pooling:max', 'Pooling:average', 'Pooling:sum']
            :param pool_seq_axis: int, default 1. Pooling axis of sequential inputs.
            :param pool_seq_l2: float, default 0.0. When pool_seq_mode is in ('Weighted', 'Attention'), it works.
            :param pool_seq_initializer: Optional[str], default None. When None, activation judge first,
                                he_uniform end. When pool_seq_mode is in ('Weighted', 'Attention'), it works.

        Task SampledSoftmaxLoss Parameters:
            :param sample_mode: Optional[str], default 'Sample:batch'. Union['Sample:all', 'Sample:batch', 'Sample:uniform',
                                'Sample:log_uniform', 'Sample:fixed_unigram', 'Sample:learned_unigram']
            :param sample_num: Optional[int], default None. When None, equals multiclass_num. Must <= multiclass_num.
            :param sample_item_distribution_list: Optional[list], default None. When None,
                                equals batch_item_distribution like batch_normalization.
            :param sample_fixed_unigram_frequency_list: Optional[list], default None. When None, list of same frequency.
                                When sample_mode is 'Sample:fixed_unigram', it works.
            :param sample_fixed_unigram_distortion: float, default 1.0. When sample_mode is 'Sample:fixed_unigram', it works.

        Task Effective Parameters:
            :param model_embeds_if_l2_norm: bool, default True.
            :param model_result_temperature_ratio: Optional[float], default None, When None, equals 1.0.

        Task Model Parameters:
            :param user_dnn_hidden_units: tuple, default ().
            :param item_dnn_hidden_units: tuple, default ().
            :param dnn_hidden_units: tuple, default (1024, 512).
            :param dnn_activation: Optional[str], default 'tanh'.
            :param dnn_dropout: float, default 0.0.
            :param dnn_if_bn: bool, default False. Batch Normalization.
            :param dnn_if_ln: bool, default False. Layer Normalization.
            :param dnn_l2: float, default 0.0.
            :param dnn_initializer: Optional[str], default None. When None, activation judge first, he_uniform end.
        """
        super().__init__()
        if task.startswith('multiclass:'):
            self.class_num = int(task.split(':')[1])
            self.user_dnn_hidden_units = user_dnn_hidden_units + (self.class_num,)
            self.item_dnn_hidden_units = item_dnn_hidden_units + (self.class_num,)
            self.dnn_hidden_units = dnn_hidden_units + (self.class_num,)

        if not (self.user_dnn_hidden_units[-1] == self.item_dnn_hidden_units[-1] == self.dnn_hidden_units[-1]):
            raise MLGBError
        if pool_mv_mode not in MVPoolModeList:
            raise MLGBError
        if pool_seq_mode not in MVPoolModeList:
            raise MLGBError
        if not embed_dim:
            raise MLGBError
        if model_result_temperature_ratio and not (1e-2 <= model_result_temperature_ratio <= 1.0):
            raise MLGBError
        if sample_mode and sample_mode not in SampleModeList:
            raise MLGBError

        user_feature_names, item_feature_names = feature_names

        self.user_dnn_embedding_fn = DeepStructuredSemanticModelUserEmbeddingLayer(
            user_feature_names=user_feature_names,
            user_inputs_if_multivalued=user_inputs_if_multivalued,
            user_inputs_if_sequential=user_inputs_if_sequential,
            user_inputs_if_embed_dense=True,
            embed_dim=embed_dim,
            embed_l2=embed_l2,
            embed_initializer=embed_initializer,
            pool_mv_mode=pool_mv_mode,
            pool_mv_axis=pool_mv_axis,
            pool_mv_l2=pool_mv_l2,
            pool_mv_initializer=pool_mv_initializer,
            pool_seq_mode=pool_seq_mode,
            pool_seq_axis=pool_seq_axis,
            pool_seq_l2=pool_seq_l2,
            pool_seq_initializer=pool_seq_initializer,
            user_dnn_hidden_units=self.user_dnn_hidden_units,
            dnn_activation=None,
            tower_embeds_flatten_mode='flatten',
            tower_embeds_if_l2_norm=model_embeds_if_l2_norm,
            seed=seed,
        )
        self.item_dnn_embedding_fn = DeepStructuredSemanticModelItemEmbeddingLayer(
            item_feature_names=item_feature_names,
            item_inputs_if_multivalued=item_inputs_if_multivalued,
            item_inputs_if_sequential=False,
            item_inputs_if_embed_dense=True,
            embed_dim=embed_dim,
            embed_l2=embed_l2,
            embed_initializer=embed_initializer,
            pool_mv_mode=pool_mv_mode,
            pool_mv_axis=pool_mv_axis,
            pool_mv_l2=pool_mv_l2,
            pool_mv_initializer=pool_mv_initializer,
            pool_seq_mode=pool_seq_mode,
            pool_seq_axis=pool_seq_axis,
            pool_seq_l2=pool_seq_l2,
            pool_seq_initializer=pool_seq_initializer,
            item_dnn_hidden_units=self.item_dnn_hidden_units,
            dnn_activation=None,
            tower_embeds_flatten_mode='flatten',
            tower_embeds_if_l2_norm=model_embeds_if_l2_norm,
            seed=seed,
        )
        self.user_mf_embedding_fn = DeepStructuredSemanticModelUserEmbeddingLayer(
            user_feature_names=user_feature_names,
            user_inputs_if_multivalued=user_inputs_if_multivalued,
            user_inputs_if_sequential=user_inputs_if_sequential,
            user_inputs_if_embed_dense=True,
            embed_dim=embed_dim,
            embed_l2=embed_l2,
            embed_initializer=embed_initializer,
            pool_mv_mode=pool_mv_mode,
            pool_mv_axis=pool_mv_axis,
            pool_mv_l2=pool_mv_l2,
            pool_mv_initializer=pool_mv_initializer,
            pool_seq_mode=pool_seq_mode,
            pool_seq_axis=pool_seq_axis,
            pool_seq_l2=pool_seq_l2,
            pool_seq_initializer=pool_seq_initializer,
            user_dnn_hidden_units=self.user_dnn_hidden_units,
            dnn_activation=None,
            tower_embeds_flatten_mode='sum',
            tower_embeds_if_l2_norm=model_embeds_if_l2_norm,
            seed=seed,
        )
        self.item_mf_embedding_fn = DeepStructuredSemanticModelItemEmbeddingLayer(
            item_feature_names=item_feature_names,
            item_inputs_if_multivalued=item_inputs_if_multivalued,
            item_inputs_if_sequential=False,
            item_inputs_if_embed_dense=True,
            embed_dim=embed_dim,
            embed_l2=embed_l2,
            embed_initializer=embed_initializer,
            pool_mv_mode=pool_mv_mode,
            pool_mv_axis=pool_mv_axis,
            pool_mv_l2=pool_mv_l2,
            pool_mv_initializer=pool_mv_initializer,
            pool_seq_mode=pool_seq_mode,
            pool_seq_axis=pool_seq_axis,
            pool_seq_l2=pool_seq_l2,
            pool_seq_initializer=pool_seq_initializer,
            item_dnn_hidden_units=self.item_dnn_hidden_units,
            dnn_activation=None,
            tower_embeds_flatten_mode='sum',
            tower_embeds_if_l2_norm=model_embeds_if_l2_norm,
            seed=seed,
        )
        self.ncf_fn = NeuralCollaborativeFilteringLayer(
            dnn_hidden_units=dnn_hidden_units,
            dnn_activation=dnn_activation,
            dnn_dropout=dnn_dropout,
            dnn_if_bn=dnn_if_bn,
            dnn_if_ln=dnn_if_ln,
            dnn_l2=dnn_l2,
            dnn_initializer=dnn_initializer,
            seed=seed,
        )
        self.task_fn = TaskLayer(
            task=task,
            task_linear_if_identity=False,
            task_linear_if_weighted=True,
            task_linear_if_bias=True,
            task_linear_initializer=None,
            task_multiclass_if_project=True,
            task_multiclass_if_softmax=False if sample_mode else True,  # None vs 'Sample:all'
            task_multiclass_temperature_ratio=model_result_temperature_ratio,
            seed=seed,
        )
        if task.startswith('multiclass:') and sample_mode:
            self.sampled_softmax_loss = SampledSoftmaxLossLayer(
                sample_mode=sample_mode,
                sample_num=sample_num,
                sample_item_distribution_list=sample_item_distribution_list,
                sample_fixed_unigram_distortion=sample_fixed_unigram_distortion,
                sample_fixed_unigram_frequency_list=sample_fixed_unigram_frequency_list,
                seed=seed,
            )

    @tf.function
    def call(self, inputs):
        user_inputs, item_inputs = inputs

        user_mf_embeddings = self.user_mf_embedding_fn(user_inputs)
        item_mf_embeddings = self.item_mf_embedding_fn(item_inputs)
        user_dnn_embeddings = self.user_dnn_embedding_fn(user_inputs)
        item_dnn_embeddings = self.item_dnn_embedding_fn(item_inputs)

        fm_outputs = self.ncf_fn([user_mf_embeddings, item_mf_embeddings, user_dnn_embeddings, item_dnn_embeddings])
        outputs = self.task_fn(fm_outputs)
        return outputs


class DSSMs(tf.keras.Model):
    def __init__(self, feature_names, task='multiclass:100', seed=None,
                 user_inputs_if_multivalued=False, user_inputs_if_sequential=False, user_inputs_if_embed_dense=True,
                 item_inputs_if_multivalued=False, item_inputs_if_embed_dense=True,
                 embed_dim=32, embed_l2=0.0, embed_initializer=None,
                 pool_mv_mode='Pooling:average', pool_mv_axis=2, pool_mv_l2=0.0, pool_mv_initializer=None,
                 pool_seq_mode='Pooling:average', pool_seq_axis=1, pool_seq_l2=0.0, pool_seq_initializer=None,
                 sample_mode='Sample:batch', sample_num=None, sample_item_distribution_list=None,
                 sample_fixed_unigram_frequency_list=None, sample_fixed_unigram_distortion=1.0,
                 model_embeds_flatten_mode='flatten', model_embeds_if_l2_norm=True, model_result_temperature_ratio=None,
                 user_dnn_hidden_units=(1024, 512), item_dnn_hidden_units=(1024, 512),
                 user_dnn_activation='tanh', item_dnn_activation='tanh',
                 dnn_dropout=0.0, dnn_if_bn=False, dnn_if_ln=False, dnn_l2=0.0, dnn_initializer=None,
                 ):
        super().__init__()
        if task.startswith('multiclass:'):
            self.class_num = int(task.split(':')[1])
            self.user_dnn_hidden_units = user_dnn_hidden_units + (self.class_num,)
            self.item_dnn_hidden_units = item_dnn_hidden_units + (self.class_num,)

        if not (self.user_dnn_hidden_units[-1] == self.item_dnn_hidden_units[-1]):
            raise MLGBError
        if pool_mv_mode not in MVPoolModeList:
            raise MLGBError
        if pool_seq_mode not in MVPoolModeList:
            raise MLGBError
        if not embed_dim:
            raise MLGBError
        if model_embeds_flatten_mode not in ('flatten', 'sum'):
            raise MLGBError
        if model_result_temperature_ratio and not (1e-2 <= model_result_temperature_ratio <= 1.0):
            raise MLGBError
        if sample_mode and sample_mode not in SampleModeList:
            raise MLGBError

        user_feature_names, item_feature_names = feature_names

        self.user_embedding_fn = DeepStructuredSemanticModelUserEmbeddingLayer(
            user_feature_names=user_feature_names,
            user_inputs_if_multivalued=user_inputs_if_multivalued,
            user_inputs_if_sequential=user_inputs_if_sequential,
            user_inputs_if_embed_dense=user_inputs_if_embed_dense,
            embed_dim=embed_dim,
            embed_l2=embed_l2,
            embed_initializer=embed_initializer,
            pool_mv_mode=pool_mv_mode,
            pool_mv_axis=pool_mv_axis,
            pool_mv_l2=pool_mv_l2,
            pool_mv_initializer=pool_mv_initializer,
            pool_seq_mode=pool_seq_mode,
            pool_seq_axis=pool_seq_axis,
            pool_seq_l2=pool_seq_l2,
            pool_seq_initializer=pool_seq_initializer,
            user_dnn_hidden_units=self.user_dnn_hidden_units,
            dnn_activation=user_dnn_activation,
            dnn_dropout=dnn_dropout,
            dnn_if_bn=dnn_if_bn,
            dnn_if_ln=dnn_if_ln,
            dnn_l2=dnn_l2,
            dnn_initializer=dnn_initializer,
            tower_embeds_flatten_mode=model_embeds_flatten_mode,
            tower_embeds_if_l2_norm=model_embeds_if_l2_norm,
            seed=seed,
        )
        self.item_embedding_fn = DeepStructuredSemanticModelItemEmbeddingLayer(
            item_feature_names=item_feature_names,
            item_inputs_if_multivalued=item_inputs_if_multivalued,
            item_inputs_if_sequential=False,
            item_inputs_if_embed_dense=item_inputs_if_embed_dense,
            embed_dim=embed_dim,
            embed_l2=embed_l2,
            embed_initializer=embed_initializer,
            pool_mv_mode=pool_mv_mode,
            pool_mv_axis=pool_mv_axis,
            pool_mv_l2=pool_mv_l2,
            pool_mv_initializer=pool_mv_initializer,
            pool_seq_mode=pool_seq_mode,
            pool_seq_axis=pool_seq_axis,
            pool_seq_l2=pool_seq_l2,
            pool_seq_initializer=pool_seq_initializer,
            item_dnn_hidden_units=self.item_dnn_hidden_units,
            dnn_activation=item_dnn_activation,
            dnn_dropout=dnn_dropout,
            dnn_if_bn=dnn_if_bn,
            dnn_if_ln=dnn_if_ln,
            dnn_l2=dnn_l2,
            dnn_initializer=dnn_initializer,
            tower_embeds_flatten_mode=model_embeds_flatten_mode,
            tower_embeds_if_l2_norm=model_embeds_if_l2_norm,
            seed=seed,
        )
        self.task_fn = TaskLayer(
            task=task,
            task_linear_if_identity=False,
            task_linear_if_weighted=False,
            task_linear_if_bias=False,
            task_linear_initializer=None,
            task_multiclass_if_project=False,
            task_multiclass_if_softmax=False if sample_mode else True,  # None vs 'Sample:all'
            task_multiclass_temperature_ratio=model_result_temperature_ratio,
            seed=seed,
        )
        if task.startswith('multiclass:') and sample_mode:
            self.sampled_softmax_loss = SampledSoftmaxLossLayer(
                sample_mode=sample_mode,
                sample_num=sample_num,
                sample_item_distribution_list=sample_item_distribution_list,
                sample_fixed_unigram_distortion=sample_fixed_unigram_distortion,
                sample_fixed_unigram_frequency_list=sample_fixed_unigram_frequency_list,
                seed=seed,
            )

    @tf.function
    def call(self, inputs):
        user_inputs, item_inputs = inputs

        user_embeddings = self.user_embedding_fn(user_inputs)
        item_embeddings = self.item_embedding_fn(item_inputs)

        dssm_outputs = user_embeddings * item_embeddings
        outputs = self.task_fn(dssm_outputs)
        return outputs


class DSSM(DSSMs):
    def __init__(self, feature_names, task='multiclass:100', seed=None,
                 user_inputs_if_multivalued=False, user_inputs_if_sequential=False,
                 item_inputs_if_multivalued=False,
                 embed_dim=32, embed_l2=0.0, embed_initializer=None,
                 pool_mv_mode='Pooling:average', pool_mv_axis=2, pool_mv_l2=0.0, pool_mv_initializer=None,
                 pool_seq_mode='Pooling:average', pool_seq_axis=1, pool_seq_l2=0.0, pool_seq_initializer=None,
                 sample_mode='Sample:batch', sample_num=None, sample_item_distribution_list=None,
                 sample_fixed_unigram_frequency_list=None, sample_fixed_unigram_distortion=1.0,
                 model_embeds_flatten_mode='flatten', model_embeds_if_l2_norm=True, model_result_temperature_ratio=None,
                 user_dnn_hidden_units=(1024, 512), item_dnn_hidden_units=(1024, 512),
                 user_dnn_activation='tanh', item_dnn_activation='tanh',
                 dnn_dropout=0.0, dnn_if_bn=False, dnn_if_ln=False, dnn_l2=0.0, dnn_initializer=None,
                 ):
        """
        Model Name: DSSM(DeepStructuredSemanticModel)
        Paper Team: Microsoft
        Paper Year: 2013
        Paper Name: <Learning deep structured semantic models for web search using clickthrough data>
        Paper Link: https://www.microsoft.com/en-us/research/wp-content/uploads/2016/02/cikm2013_DSSM_fullversion.pdf

        Task Inputs Parameters:
            :param feature_names: tuple(tuple(dict)), must. Embedding need vocabulary size and custom embed_dim of features.
            :param task: str, default 'multiclass:100'. Union['binary', 'regression', 'multiclass:{int}']
            :param seed: Optional[int], default None.
            :param user_inputs_if_multivalued: bool, default False.
            :param user_inputs_if_sequential: bool, default False.
            :param item_inputs_if_multivalued: bool, default False.
            :param embed_dim: int, default 32.
            :param embed_l2: float, default 0.0.
            :param embed_initializer: Optional[str], default None. When None, activation judge first, he_uniform end.
            :param pool_mv_mode: str, default 'Pooling:average'. Pooling mode of multivalued inputs. Union[
                                'Attention', 'Weighted', 'Pooling:max', 'Pooling:average', 'Pooling:sum']
            :param pool_mv_axis: int, default 2. Pooling axis of multivalued inputs.
            :param pool_mv_l2: float, default 0.0. When pool_mv_mode is in ('Weighted', 'Attention'), it works.
            :param pool_mv_initializer: Optional[str], default None. When None, activation judge first,
                                he_uniform end. When pool_mv_mode is in ('Weighted', 'Attention'), it works.
            :param pool_seq_mode: str, default 'Pooling:average'. Pooling mode of sequential inputs. Union[
                                'Attention', 'Weighted', 'Pooling:max', 'Pooling:average', 'Pooling:sum']
            :param pool_seq_axis: int, default 1. Pooling axis of sequential inputs.
            :param pool_seq_l2: float, default 0.0. When pool_seq_mode is in ('Weighted', 'Attention'), it works.
            :param pool_seq_initializer: Optional[str], default None. When None, activation judge first,
                                he_uniform end. When pool_seq_mode is in ('Weighted', 'Attention'), it works.

        Task SampledSoftmaxLoss Parameters:
            :param sample_mode: Optional[str], default 'Sample:batch'. Union['Sample:all', 'Sample:batch', 'Sample:uniform',
                                'Sample:log_uniform', 'Sample:fixed_unigram', 'Sample:learned_unigram']
            :param sample_num: Optional[int], default None. When None, equals multiclass_num. Must <= multiclass_num.
            :param sample_item_distribution_list: Optional[list], default None. When None,
                                equals batch_item_distribution like batch_normalization.
            :param sample_fixed_unigram_frequency_list: Optional[list], default None. When None, list of same frequency.
                                When sample_mode is 'Sample:fixed_unigram', it works.
            :param sample_fixed_unigram_distortion: float, default 1.0. When sample_mode is 'Sample:fixed_unigram', it works.

        Task Effective Parameters:
            :param model_embeds_flatten_mode: str, default 'flatten', Union['flatten']
            :param model_embeds_if_l2_norm: bool, default True.
            :param model_result_temperature_ratio: Optional[float], default None, When None, equals 1.0.

        Task Model Parameters:
            :param user_dnn_hidden_units: tuple, default (1024, 512).
            :param item_dnn_hidden_units: tuple, default (1024, 512).
            :param user_dnn_activation: Optional[str], default 'tanh'.
            :param user_dnn_activation: Optional[str], default 'tanh'.
            :param dnn_dropout: float, default 0.0.
            :param dnn_if_bn: bool, default False. Batch Normalization.
            :param dnn_if_ln: bool, default False. Layer Normalization.
            :param dnn_l2: float, default 0.0.
            :param dnn_initializer: Optional[str], default None. When None, activation judge first, he_uniform end.
        """
        super().__init__(
            feature_names, task, seed,
            user_inputs_if_multivalued, user_inputs_if_sequential, True,
            item_inputs_if_multivalued, True,
            embed_dim, embed_l2, embed_initializer,
            pool_mv_mode, pool_mv_axis, pool_mv_l2, pool_mv_initializer,
            pool_seq_mode, pool_seq_axis, pool_seq_l2, pool_seq_initializer,
            sample_mode, sample_num, sample_item_distribution_list,
            sample_fixed_unigram_frequency_list, sample_fixed_unigram_distortion,
            model_embeds_flatten_mode, model_embeds_if_l2_norm, model_result_temperature_ratio,
            user_dnn_hidden_units, item_dnn_hidden_units,
            user_dnn_activation, item_dnn_activation,
            dnn_dropout, dnn_if_bn, dnn_if_ln, dnn_l2, dnn_initializer,
        )
        if not (user_dnn_activation and item_dnn_activation):  # must both NN.
            raise MLGBError
        if model_embeds_flatten_mode != 'flatten':
            raise MLGBError


class EBR(DSSMs):
    def __init__(self, feature_names, task='multiclass:100', seed=None,
                 user_inputs_if_multivalued=False, user_inputs_if_sequential=False,
                 item_inputs_if_multivalued=False,
                 embed_dim=32, embed_l2=0.0, embed_initializer=None,
                 pool_mv_mode='Pooling:average', pool_mv_axis=2, pool_mv_l2=0.0, pool_mv_initializer=None,
                 pool_seq_mode='Pooling:average', pool_seq_axis=1, pool_seq_l2=0.0, pool_seq_initializer=None,
                 sample_mode='Sample:batch', sample_num=None, sample_item_distribution_list=None,
                 sample_fixed_unigram_frequency_list=None, sample_fixed_unigram_distortion=1.0,
                 model_embeds_flatten_mode='flatten', model_embeds_if_l2_norm=True, model_result_temperature_ratio=None,
                 user_dnn_hidden_units=(), item_dnn_hidden_units=(),
                 user_dnn_activation=None, item_dnn_activation=None,  #
                 dnn_dropout=0.0, dnn_if_bn=False, dnn_if_ln=False, dnn_l2=0.0, dnn_initializer=None,
                 ):
        """
        Model Name: EBR(EmbeddingBasedRetrieval)
        Paper Team: Facebook(Meta), etc.
        Paper Year: 2020
        Paper Name: <Embedding-based Retrieval in Facebook Search>
        Paper Link: https://browse.arxiv.org/pdf/2006.11632.pdf

        Task Inputs Parameters:
            :param feature_names: tuple(tuple(dict)), must. Embedding need vocabulary size and custom embed_dim of features.
            :param task: str, default 'multiclass:100'. Union['binary', 'regression', 'multiclass:{int}']
            :param seed: Optional[int], default None.
            :param user_inputs_if_multivalued: bool, default False.
            :param user_inputs_if_sequential: bool, default False.
            :param item_inputs_if_multivalued: bool, default False.
            :param embed_dim: int, default 32.
            :param embed_l2: float, default 0.0.
            :param embed_initializer: Optional[str], default None. When None, activation judge first, he_uniform end.
            :param pool_mv_mode: str, default 'Pooling:average'. Pooling mode of multivalued inputs. Union[
                                'Attention', 'Weighted', 'Pooling:max', 'Pooling:average', 'Pooling:sum']
            :param pool_mv_axis: int, default 2. Pooling axis of multivalued inputs.
            :param pool_mv_l2: float, default 0.0. When pool_mv_mode is in ('Weighted', 'Attention'), it works.
            :param pool_mv_initializer: Optional[str], default None. When None, activation judge first,
                                he_uniform end. When pool_mv_mode is in ('Weighted', 'Attention'), it works.
            :param pool_seq_mode: str, default 'Pooling:average'. Pooling mode of sequential inputs. Union[
                                'Attention', 'Weighted', 'Pooling:max', 'Pooling:average', 'Pooling:sum']
            :param pool_seq_axis: int, default 1. Pooling axis of sequential inputs.
            :param pool_seq_l2: float, default 0.0. When pool_seq_mode is in ('Weighted', 'Attention'), it works.
            :param pool_seq_initializer: Optional[str], default None. When None, activation judge first,
                                he_uniform end. When pool_seq_mode is in ('Weighted', 'Attention'), it works.

        Task SampledSoftmaxLoss Parameters:
            :param sample_mode: Optional[str], default 'Sample:batch'. Union['Sample:all', 'Sample:batch', 'Sample:uniform',
                                'Sample:log_uniform', 'Sample:fixed_unigram', 'Sample:learned_unigram']
            :param sample_num: Optional[int], default None. When None, equals multiclass_num. Must <= multiclass_num.
            :param sample_item_distribution_list: Optional[list], default None. When None,
                                equals batch_item_distribution like batch_normalization.
            :param sample_fixed_unigram_frequency_list: Optional[list], default None. When None, list of same frequency.
                                When sample_mode is 'Sample:fixed_unigram', it works.
            :param sample_fixed_unigram_distortion: float, default 1.0. When sample_mode is 'Sample:fixed_unigram', it works.

        Task Effective Parameters:
            :param model_embeds_flatten_mode: str, default 'flatten', Union['flatten']
            :param model_embeds_if_l2_norm: bool, default True.
            :param model_result_temperature_ratio: Optional[float], default None, When None, equals 1.0.

        Task Model Parameters:
            :param user_dnn_hidden_units: tuple, default ().
            :param item_dnn_hidden_units: tuple, default ().
            :param user_dnn_activation: Optional[str], default None.
            :param user_dnn_activation: Optional[str], default None.
            :param dnn_dropout: float, default 0.0.
            :param dnn_if_bn: bool, default False. Batch Normalization.
            :param dnn_if_ln: bool, default False. Layer Normalization.
            :param dnn_l2: float, default 0.0.
            :param dnn_initializer: Optional[str], default None. When None, activation judge first, he_uniform end.
        """
        super().__init__(
            feature_names, task, seed,
            user_inputs_if_multivalued, user_inputs_if_sequential, True,
            item_inputs_if_multivalued, True,
            embed_dim, embed_l2, embed_initializer,
            pool_mv_mode, pool_mv_axis, pool_mv_l2, pool_mv_initializer,
            pool_seq_mode, pool_seq_axis, pool_seq_l2, pool_seq_initializer,
            sample_mode, sample_num, sample_item_distribution_list,
            sample_fixed_unigram_frequency_list, sample_fixed_unigram_distortion,
            model_embeds_flatten_mode, model_embeds_if_l2_norm, model_result_temperature_ratio,
            user_dnn_hidden_units, item_dnn_hidden_units,
            user_dnn_activation, item_dnn_activation,
            dnn_dropout, dnn_if_bn, dnn_if_ln, dnn_l2, dnn_initializer,
        )
        if user_dnn_activation or item_dnn_activation:  # only projection, not NN.
            raise MLGBError
        if model_embeds_flatten_mode != 'flatten':
            raise MLGBError


class MatchFM(DSSMs):
    def __init__(self, feature_names, task='multiclass:100', seed=None,
                 user_inputs_if_multivalued=False, user_inputs_if_sequential=False,
                 item_inputs_if_multivalued=False,
                 embed_dim=32, embed_l2=0.0, embed_initializer=None,
                 pool_mv_mode='Pooling:average', pool_mv_axis=2, pool_mv_l2=0.0, pool_mv_initializer=None,
                 pool_seq_mode='Pooling:average', pool_seq_axis=1, pool_seq_l2=0.0, pool_seq_initializer=None,
                 sample_mode='Sample:batch', sample_num=None, sample_item_distribution_list=None,
                 sample_fixed_unigram_frequency_list=None, sample_fixed_unigram_distortion=1.0,
                 model_embeds_flatten_mode='sum', model_embeds_if_l2_norm=True, model_result_temperature_ratio=None,
                 user_dnn_hidden_units=(), item_dnn_hidden_units=(),
                 user_dnn_activation=None, item_dnn_activation=None,  #
                 dnn_dropout=0.0, dnn_if_bn=False, dnn_if_ln=False, dnn_l2=0.0, dnn_initializer=None,
                 ):
        """
        Model Name: FM(FactorizationMachine)
        Paper Team: Steffen Rendle(Google, 2013-Present)
        Paper Year: 2010
        Paper Name: <Factorization Machines>
        Paper Link: https://www.csie.ntu.edu.tw/~b97053/paper/Rendle2010FM.pdf,
                    https://cseweb.ucsd.edu/classes/fa17/cse291-b/reading/Rendle2010FM.pdf,
                    https://analyticsconsultores.com.mx/wp-content/uploads/2019/03/Factorization-Machines-Steffen-Rendle-Osaka-University-2010.pdf

        Task Inputs Parameters:
            :param feature_names: tuple(tuple(dict)), must. Embedding need vocabulary size and custom embed_dim of features.
            :param task: str, default 'multiclass:100'. Union['binary', 'regression', 'multiclass:{int}']
            :param seed: Optional[int], default None.
            :param user_inputs_if_multivalued: bool, default False.
            :param user_inputs_if_sequential: bool, default False.
            :param item_inputs_if_multivalued: bool, default False.
            :param embed_dim: int, default 32.
            :param embed_l2: float, default 0.0.
            :param embed_initializer: Optional[str], default None. When None, activation judge first, he_uniform end.
            :param pool_mv_mode: str, default 'Pooling:average'. Pooling mode of multivalued inputs. Union[
                                'Attention', 'Weighted', 'Pooling:max', 'Pooling:average', 'Pooling:sum']
            :param pool_mv_axis: int, default 2. Pooling axis of multivalued inputs.
            :param pool_mv_l2: float, default 0.0. When pool_mv_mode is in ('Weighted', 'Attention'), it works.
            :param pool_mv_initializer: Optional[str], default None. When None, activation judge first,
                                he_uniform end. When pool_mv_mode is in ('Weighted', 'Attention'), it works.
            :param pool_seq_mode: str, default 'Pooling:average'. Pooling mode of sequential inputs. Union[
                                'Attention', 'Weighted', 'Pooling:max', 'Pooling:average', 'Pooling:sum']
            :param pool_seq_axis: int, default 1. Pooling axis of sequential inputs.
            :param pool_seq_l2: float, default 0.0. When pool_seq_mode is in ('Weighted', 'Attention'), it works.
            :param pool_seq_initializer: Optional[str], default None. When None, activation judge first,
                                he_uniform end. When pool_seq_mode is in ('Weighted', 'Attention'), it works.

        Task SampledSoftmaxLoss Parameters:
            :param sample_mode: Optional[str], default 'Sample:batch'. Union['Sample:all', 'Sample:batch', 'Sample:uniform',
                                'Sample:log_uniform', 'Sample:fixed_unigram', 'Sample:learned_unigram']
            :param sample_num: Optional[int], default None. When None, equals multiclass_num. Must <= multiclass_num.
            :param sample_item_distribution_list: Optional[list], default None. When None,
                                equals batch_item_distribution like batch_normalization.
            :param sample_fixed_unigram_frequency_list: Optional[list], default None. When None, list of same frequency.
                                When sample_mode is 'Sample:fixed_unigram', it works.
            :param sample_fixed_unigram_distortion: float, default 1.0. When sample_mode is 'Sample:fixed_unigram', it works.

        Task Effective Parameters:
            :param model_embeds_flatten_mode: str, default 'sum', Union['sum']
            :param model_embeds_if_l2_norm: bool, default True.
            :param model_result_temperature_ratio: Optional[float], default None, When None, equals 1.0.

        Task Model Parameters:
            :param user_dnn_hidden_units: tuple, default ().
            :param item_dnn_hidden_units: tuple, default ().
            :param user_dnn_activation: Optional[str], default None.
            :param user_dnn_activation: Optional[str], default None.
            :param dnn_dropout: float, default 0.0.
            :param dnn_if_bn: bool, default False. Batch Normalization.
            :param dnn_if_ln: bool, default False. Layer Normalization.
            :param dnn_l2: float, default 0.0.
            :param dnn_initializer: Optional[str], default None. When None, activation judge first, he_uniform end.
        """
        super().__init__(
            feature_names, task, seed,
            user_inputs_if_multivalued, user_inputs_if_sequential, True,
            item_inputs_if_multivalued, True,
            embed_dim, embed_l2, embed_initializer,
            pool_mv_mode, pool_mv_axis, pool_mv_l2, pool_mv_initializer,
            pool_seq_mode, pool_seq_axis, pool_seq_l2, pool_seq_initializer,
            sample_mode, sample_num, sample_item_distribution_list,
            sample_fixed_unigram_frequency_list, sample_fixed_unigram_distortion,
            model_embeds_flatten_mode, model_embeds_if_l2_norm, model_result_temperature_ratio,
            user_dnn_hidden_units, item_dnn_hidden_units,
            user_dnn_activation, item_dnn_activation,
            dnn_dropout, dnn_if_bn, dnn_if_ln, dnn_l2, dnn_initializer,
        )
        if user_dnn_activation or item_dnn_activation:  # only projection, not NN.
            raise MLGBError
        if model_embeds_flatten_mode != 'sum':
            raise MLGBError


class YoutubeDNN(DSSMs):
    def __init__(self, feature_names, task='multiclass:100', seed=None,
                 user_inputs_if_multivalued=False, user_inputs_if_sequential=False,
                 embed_dim=32, embed_l2=0.0, embed_initializer=None,
                 pool_mv_mode='Pooling:average', pool_mv_axis=2, pool_mv_l2=0.0, pool_mv_initializer=None,
                 pool_seq_mode='Pooling:average', pool_seq_axis=1, pool_seq_l2=0.0, pool_seq_initializer=None,
                 sample_mode='Sample:batch', sample_num=None, sample_item_distribution_list=None,
                 sample_fixed_unigram_frequency_list=None, sample_fixed_unigram_distortion=1.0,
                 model_embeds_flatten_mode='flatten', model_embeds_if_l2_norm=True, model_result_temperature_ratio=None,
                 user_dnn_hidden_units=(1024, 512), item_dnn_hidden_units=(),  #
                 user_dnn_activation='tanh', item_dnn_activation=None,  #
                 dnn_dropout=0.0, dnn_if_bn=False, dnn_if_ln=False, dnn_l2=0.0, dnn_initializer=None,
                 ):
        """
        Model Name: YoutubeDNN(DeepNeuralNetworkForYoutube)
        Paper Team: Google(Alphabet)
        Paper Year: 2016
        Paper Name: <Deep Neural Networks for YouTube Recommendations>
        Paper Link: https://static.googleusercontent.com/media/research.google.com/zh-CN//pubs/archive/45530.pdf,
                    https://dl.acm.org/doi/pdf/10.1145/2959100.2959190

        Task Inputs Parameters:
            :param feature_names: tuple(tuple(dict)), must. Embedding need vocabulary size and custom embed_dim of features.
            :param task: str, default 'multiclass:100'. Union['binary', 'regression', 'multiclass:{int}']
            :param seed: Optional[int], default None.
            :param user_inputs_if_multivalued: bool, default False.
            :param user_inputs_if_sequential: bool, default False.
            :param embed_dim: int, default 32.
            :param embed_l2: float, default 0.0.
            :param embed_initializer: Optional[str], default None. When None, activation judge first, he_uniform end.
            :param pool_mv_mode: str, default 'Pooling:average'. Pooling mode of multivalued inputs. Union[
                                'Attention', 'Weighted', 'Pooling:max', 'Pooling:average', 'Pooling:sum']
            :param pool_mv_axis: int, default 2. Pooling axis of multivalued inputs.
            :param pool_mv_l2: float, default 0.0. When pool_mv_mode is in ('Weighted', 'Attention'), it works.
            :param pool_mv_initializer: Optional[str], default None. When None, activation judge first,
                                he_uniform end. When pool_mv_mode is in ('Weighted', 'Attention'), it works.
            :param pool_seq_mode: str, default 'Pooling:average'. Pooling mode of sequential inputs. Union[
                                'Attention', 'Weighted', 'Pooling:max', 'Pooling:average', 'Pooling:sum']
            :param pool_seq_axis: int, default 1. Pooling axis of sequential inputs.
            :param pool_seq_l2: float, default 0.0. When pool_seq_mode is in ('Weighted', 'Attention'), it works.
            :param pool_seq_initializer: Optional[str], default None. When None, activation judge first,
                                he_uniform end. When pool_seq_mode is in ('Weighted', 'Attention'), it works.

        Task SampledSoftmaxLoss Parameters:
            :param sample_mode: Optional[str], default 'Sample:batch'. Union['Sample:all', 'Sample:batch', 'Sample:uniform',
                                'Sample:log_uniform', 'Sample:fixed_unigram', 'Sample:learned_unigram']
            :param sample_num: Optional[int], default None. When None, equals multiclass_num. Must <= multiclass_num.
            :param sample_item_distribution_list: Optional[list], default None. When None,
                                equals batch_item_distribution like batch_normalization.
            :param sample_fixed_unigram_frequency_list: Optional[list], default None. When None, list of same frequency.
                                When sample_mode is 'Sample:fixed_unigram', it works.
            :param sample_fixed_unigram_distortion: float, default 1.0. When sample_mode is 'Sample:fixed_unigram', it works.

        Task Effective Parameters:
            :param model_embeds_flatten_mode: str, default 'flatten', Union['flatten']
            :param model_embeds_if_l2_norm: bool, default True.
            :param model_result_temperature_ratio: Optional[float], default None, When None, equals 1.0.

        Task Model Parameters:
            :param user_dnn_hidden_units: tuple, default (1024, 512).
            :param item_dnn_hidden_units: tuple, default ().
            :param user_dnn_activation: Optional[str], default 'tanh'.
            :param item_dnn_activation: Optional[str], default None.
            :param dnn_dropout: float, default 0.0.
            :param dnn_if_bn: bool, default False. Batch Normalization.
            :param dnn_if_ln: bool, default False. Layer Normalization.
            :param dnn_l2: float, default 0.0.
            :param dnn_initializer: Optional[str], default None. When None, activation judge first, he_uniform end.
        """
        super().__init__(
            feature_names, task, seed,
            user_inputs_if_multivalued, user_inputs_if_sequential, True,
            False, False,  # Only need item_id embeddings.
            embed_dim, embed_l2, embed_initializer,
            pool_mv_mode, pool_mv_axis, pool_mv_l2, pool_mv_initializer,
            pool_seq_mode, pool_seq_axis, pool_seq_l2, pool_seq_initializer,
            sample_mode, sample_num, sample_item_distribution_list,
            sample_fixed_unigram_frequency_list, sample_fixed_unigram_distortion,
            model_embeds_flatten_mode, model_embeds_if_l2_norm, model_result_temperature_ratio,
            user_dnn_hidden_units, item_dnn_hidden_units,
            user_dnn_activation, item_dnn_activation,
            dnn_dropout, dnn_if_bn, dnn_if_ln, dnn_l2, dnn_initializer,
        )
        if item_dnn_activation:  # only projection, not NN.
            raise MLGBError
        if model_embeds_flatten_mode != 'flatten':
            raise MLGBError


class MIND(tf.keras.Model):
    def __init__(self, feature_names, task='multiclass:100', seed=None,
                 user_inputs_if_multivalued=False, user_inputs_if_sequential=True,  #
                 embed_dim=32, embed_l2=0.0, embed_initializer=None,
                 pool_mv_mode='Pooling:average', pool_mv_axis=2, pool_mv_l2=0.0, pool_mv_initializer=None,
                 pool_seq_mode='Pooling:average', pool_seq_axis=1, pool_seq_l2=0.0, pool_seq_initializer=None,
                 sample_mode='Sample:batch', sample_num=None, sample_item_distribution_list=None,
                 sample_fixed_unigram_frequency_list=None, sample_fixed_unigram_distortion=1.0,
                 model_embeds_if_l2_norm=True, model_result_temperature_ratio=None,
                 mind_if_label_attention=False,
                 user_dnn_hidden_units=(1024, 512), item_dnn_hidden_units=(),  #
                 capsule_num=3, capsule_activation='squash', capsule_l2=0.0, capsule_initializer=None,
                 capsule_interest_num_if_dynamic=False, capsule_input_sequence_pad_mode='pre',
                 capsule_routing_initializer='random_normal',
                 dnn_activation='tanh', dnn_dropout=0.0, dnn_if_bn=False, dnn_if_ln=False, dnn_l2=0.0, dnn_initializer=None,
                 ):
        """
        Model Name: MIND(MultiInterestNetworkWithDynamicRouting)
        Paper Team: Alibaba
        Paper Year: 2019
        Paper Name: <Multi-Interest Network with Dynamic Routing for Recommendation at Tmall>
        Paper Link: https://arxiv.org/pdf/1904.08030.pdf

        Task Inputs Parameters:
            :param feature_names: tuple(tuple(dict)), must. Embedding need vocabulary size and custom embed_dim of features.
            :param task: str, default 'multiclass:100'. Union['binary', 'regression', 'multiclass:{int}']
            :param seed: Optional[int], default None.
            :param user_inputs_if_multivalued: bool, default False.
            :param user_inputs_if_sequential: bool, default False.
            :param embed_dim: int, default 32.
            :param embed_l2: float, default 0.0.
            :param embed_initializer: Optional[str], default None. When None, activation judge first, he_uniform end.
            :param pool_mv_mode: str, default 'Pooling:average'. Pooling mode of multivalued inputs. Union[
                                'Attention', 'Weighted', 'Pooling:max', 'Pooling:average', 'Pooling:sum']
            :param pool_mv_axis: int, default 2. Pooling axis of multivalued inputs.
            :param pool_mv_l2: float, default 0.0. When pool_mv_mode is in ('Weighted', 'Attention'), it works.
            :param pool_mv_initializer: Optional[str], default None. When None, activation judge first,
                                he_uniform end. When pool_mv_mode is in ('Weighted', 'Attention'), it works.
            :param pool_seq_mode: str, default 'Pooling:average'. Pooling mode of sequential inputs. Union[
                                'Attention', 'Weighted', 'Pooling:max', 'Pooling:average', 'Pooling:sum']
            :param pool_seq_axis: int, default 1. Pooling axis of sequential inputs.
            :param pool_seq_l2: float, default 0.0. When pool_seq_mode is in ('Weighted', 'Attention'), it works.
            :param pool_seq_initializer: Optional[str], default None. When None, activation judge first,
                                he_uniform end. When pool_seq_mode is in ('Weighted', 'Attention'), it works.

        Task SampledSoftmaxLoss Parameters:
            :param sample_mode: Optional[str], default 'Sample:batch'. Union['Sample:all', 'Sample:batch', 'Sample:uniform',
                                'Sample:log_uniform', 'Sample:fixed_unigram', 'Sample:learned_unigram']
            :param sample_num: Optional[int], default None. When None, equals multiclass_num. Must <= multiclass_num.
            :param sample_item_distribution_list: Optional[list], default None. When None,
                                equals batch_item_distribution like batch_normalization.
            :param sample_fixed_unigram_frequency_list: Optional[list], default None. When None, list of same frequency.
                                When sample_mode is 'Sample:fixed_unigram', it works.
            :param sample_fixed_unigram_distortion: float, default 1.0. When sample_mode is 'Sample:fixed_unigram', it works.

        Task Effective Parameters:
            :param model_embeds_if_l2_norm: bool, default True.
            :param model_result_temperature_ratio: Optional[float], default None, When None, equals 1.0.

        Task Model Parameters:
            :param mind_if_label_attention: bool, default False.
            :param user_dnn_hidden_units: tuple, default (1024, 512).
            :param item_dnn_hidden_units: tuple, default ().
            :param capsule_num: int, default 3.
            :param capsule_activation: Optional[str], default 'squash'.
            :param capsule_l2: float, default 0.0.
            :param capsule_initializer: Optional[str], default None. When None, activation judge first, he_uniform end.
            :param capsule_interest_num_if_dynamic: bool, default False.
            :param capsule_input_sequence_pad_mode: str, default 'pre'.
            :param capsule_routing_initializer: Optional[str], default 'random_normal'. When None, activation judge first, he_uniform end.
            :param dnn_activation: Optional[str], default 'tanh'.
            :param dnn_if_bn: bool, default False. Batch Normalization.
            :param dnn_if_ln: bool, default False. Layer Normalization.
            :param dnn_dropout: float, default 0.0.
            :param dnn_l2: float, default 0.0.
            :param dnn_initializer: Optional[str], default None. When None, activation judge first, he_uniform end.
        """
        super().__init__()
        if task.startswith('multiclass:'):
            self.class_num = int(task.split(':')[1])
            self.user_dnn_hidden_units = user_dnn_hidden_units + (self.class_num,)
            self.item_dnn_hidden_units = item_dnn_hidden_units + (self.class_num,)

        if not (self.user_dnn_hidden_units[-1] == self.item_dnn_hidden_units[-1]):
            raise MLGBError
        if not user_inputs_if_sequential:
            raise MLGBError
        if capsule_input_sequence_pad_mode not in ('pre', 'post'):
            raise MLGBError
        if pool_mv_mode not in MVPoolModeList:
            raise MLGBError
        if pool_seq_mode not in MVPoolModeList:
            raise MLGBError
        if not embed_dim:
            raise MLGBError
        if model_result_temperature_ratio and not (1e-2 <= model_result_temperature_ratio <= 1.0):
            raise MLGBError
        if sample_mode and sample_mode not in SampleModeList:
            raise MLGBError

        user_feature_names, item_feature_names = feature_names
        self.mind_if_label_attention = mind_if_label_attention

        self.user_embedding_fn = UserInputs_MultiInterestNetworkWithDynamicRoutingLayer(
            user_feature_names=user_feature_names,
            user_inputs_if_multivalued=user_inputs_if_multivalued,
            user_inputs_if_sequential=user_inputs_if_sequential,  #
            embed_dim=embed_dim,
            embed_l2=embed_l2,
            embed_initializer=embed_initializer,
            pool_mv_mode=pool_mv_mode,
            pool_mv_axis=pool_mv_axis,
            pool_mv_l2=pool_mv_l2,
            pool_mv_initializer=pool_mv_initializer,
            pool_seq_mode=pool_seq_mode,
            pool_seq_axis=pool_seq_axis,  # (b, s, e)
            pool_seq_l2=pool_seq_l2,
            pool_seq_initializer=pool_seq_initializer,
            tower_embeds_if_l2_norm=model_embeds_if_l2_norm,  #
            capsule_num=capsule_num,
            capsule_activation=capsule_activation,
            capsule_l2=capsule_l2,
            capsule_initializer=capsule_initializer,
            capsule_interest_num_if_dynamic=capsule_interest_num_if_dynamic,
            capsule_input_sequence_pad_mode=capsule_input_sequence_pad_mode,
            capsule_routing_initializer=capsule_routing_initializer,
            dnn_hidden_units=self.user_dnn_hidden_units,
            dnn_activation=dnn_activation,
            dnn_dropout=dnn_dropout,
            dnn_if_bn=dnn_if_bn,
            dnn_if_ln=dnn_if_ln,
            dnn_l2=dnn_l2,
            dnn_initializer=dnn_initializer,
            seed=seed,
        )
        self.item_embedding_fn = DeepStructuredSemanticModelItemEmbeddingLayer(
            item_feature_names=item_feature_names,
            item_inputs_if_multivalued=False,
            item_inputs_if_sequential=False,
            item_inputs_if_embed_dense=False,  # Only need item_id embeddings.
            embed_dim=embed_dim,
            embed_l2=embed_l2,
            embed_initializer=embed_initializer,
            pool_mv_mode=pool_mv_mode,
            pool_mv_axis=pool_mv_axis,
            pool_mv_l2=pool_mv_l2,
            pool_mv_initializer=pool_mv_initializer,
            pool_seq_mode=pool_seq_mode,
            pool_seq_axis=pool_seq_axis,
            pool_seq_l2=pool_seq_l2,
            pool_seq_initializer=pool_seq_initializer,
            item_dnn_hidden_units=self.item_dnn_hidden_units,
            dnn_activation=None,
            tower_embeds_flatten_mode='flatten',
            tower_embeds_if_l2_norm=model_embeds_if_l2_norm,
            seed=seed,
        )
        if self.mind_if_label_attention:
            self.label_attention_fn = LabelAttentionLayer(
                softmax_axis=1,
                softmax_pre_temperature_ratio=model_result_temperature_ratio if model_result_temperature_ratio else 1.0,
            )
        self.task_fn = TaskLayer(
            task=task,
            task_linear_if_identity=False,
            task_linear_if_weighted=False,
            task_linear_if_bias=False,
            task_linear_initializer=None,
            task_multiclass_if_project=False,
            task_multiclass_if_softmax=False if sample_mode else True,  # None vs 'Sample:all'
            task_multiclass_temperature_ratio=model_result_temperature_ratio,
            seed=seed,
        )
        if task.startswith('multiclass:') and sample_mode:
            self.sampled_softmax_loss = SampledSoftmaxLossLayer(
                sample_mode=sample_mode,
                sample_num=sample_num,
                sample_item_distribution_list=sample_item_distribution_list,
                sample_fixed_unigram_distortion=sample_fixed_unigram_distortion,
                sample_fixed_unigram_frequency_list=sample_fixed_unigram_frequency_list,
                seed=seed,
            )

    @tf.function
    def call(self, inputs):
        user_inputs, item_inputs = inputs

        user_embeddings = self.user_embedding_fn(user_inputs)
        item_embeddings = self.item_embedding_fn(item_inputs)

        if self.mind_if_label_attention:
            mind_outputs = self.label_attention_fn([item_embeddings, user_embeddings])  # q, k=v
        else:
            mind_outputs = user_embeddings * item_embeddings
        outputs = self.task_fn(mind_outputs)
        return outputs




































# coding=utf-8
# author=uliontse

"""
Copyright 2020 UlionTse

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from mlgb.tf.configs import (
    tf,
    MVPoolModeList,
)
from mlgb.tf.inputs import InputsLayer
from mlgb.tf.components.linears import TaskLayer
from mlgb.tf.modules.ranking.multitask import (
    SharedBottomLayer,
    EntireSpaceMultitaskModelLayer,
    MultigateMixtureOfExpertLayer,
    ProgressiveLayeredExtractionLayer,
    ParameterEmbeddingPersonalizedNetworkLayer,
)
from mlgb.error import MLGBError

__all__ = [
    'SharedBottom', 'ESMM', 'MMoE', 'PLE', 'PEPNet',
]


class SharedBottom(tf.keras.Model):
    def __init__(self, feature_names, task=('binary', 'binary'), seed=None,
                 inputs_if_multivalued=False, inputs_if_sequential=False, inputs_if_embed_dense=False,
                 embed_dim=32, embed_2d_dim=None, embed_l2=0.0, embed_initializer=None,
                 pool_mv_mode='Pooling:average', pool_mv_axis=2, pool_mv_l2=0.0, pool_mv_initializer=None,
                 pool_seq_mode='Pooling:average', pool_seq_axis=1, pool_seq_l2=0.0, pool_seq_initializer=None,
                 share_mode='SB:hard',
                 soft_dnn_hidden_units=(64, 32),
                 hard_bottom_dnn_hidden_units=(64, 32), hard_tower_dnn_hidden_units=(32,),
                 dnn_activation='selu', dnn_dropout=0.0, dnn_if_bn=False, dnn_if_ln=False,
                 dnn_l2=0.0, dnn_initializer=None,
                 ):
        """
        Model Name: SharedBottom(SharedBottom)
        Paper Team: Sebastian Ruder(InsightCentre)
        Paper Year: 2017
        Paper Name: <An Overview of Multi-Task Learning in Deep Neural Networks>
        Paper Link: https://arxiv.org/pdf/1706.05098.pdf

        Task Inputs Parameters:
            :param feature_names: tuple(tuple(dict)), must. Embedding need vocabulary size and custom embed_dim of features.
            :param task: Tuple[str], default ('binary', 'binary').
            :param seed: Optional[int], default None.
            :param inputs_if_multivalued: bool, default False.
            :param inputs_if_sequential: bool, default False.
            :param inputs_if_embed_dense: bool, default False.
            :param embed_dim: int, default 32.
            :param embed_2d_dim: Optional[int], default None. When None, each field has own embed_dim by feature_names.
            :param embed_l2: float, default 0.0.
            :param embed_initializer: Optional[str], default None. When None, activation judge first, he_uniform end.
            :param pool_mv_mode: str, default 'Pooling:average'. Pooling mode of multivalued inputs. Union[
                                'Attention', 'Weighted', 'Pooling:max', 'Pooling:average', 'Pooling:sum']
            :param pool_mv_axis: int, default 2. Pooling axis of multivalued inputs.
            :param pool_mv_l2: float, default 0.0. When pool_mv_mode is in ('Weighted', 'Attention'), it works.
            :param pool_mv_initializer: Optional[str], default None. When None, activation judge first,
                                he_uniform end. When pool_mv_mode is in ('Weighted', 'Attention'), it works.
            :param pool_seq_mode: str, default 'Pooling:average'. Pooling mode of sequential inputs. Union[
                                'Attention', 'Weighted', 'Pooling:max', 'Pooling:average', 'Pooling:sum']
            :param pool_seq_axis: int, default 1. Pooling axis of sequential inputs.
            :param pool_seq_l2: float, default 0.0. When pool_seq_mode is in ('Weighted', 'Attention'), it works.
            :param pool_seq_initializer: Optional[str], default None. When None, activation judge first,
                                he_uniform end. When pool_seq_mode is in ('Weighted', 'Attention'), it works.

        Task Model Parameters:
            :param share_mode: str, default 'SB:hard'. Union['SB:hard', 'SB:soft']
            :param soft_dnn_hidden_units: Tuple[int], default (64, 32).
            :param hard_bottom_dnn_hidden_units: Tuple[int], default (64, 32).
            :param hard_tower_dnn_hidden_units: Tuple[int], default (32,).
            :param dnn_activation: Optional[str], default 'selu'.
            :param dnn_dropout: float, default 0.0.
            :param dnn_if_bn: bool, default False. Batch Normalization.
            :param dnn_if_ln: bool, default False. Layer Normalization.
            :param dnn_l2: float, default 0.0.
            :param dnn_initializer: Optional[str], default None. When None, activation judge first, he_uniform end.
        """
        super().__init__()
        task_list = task
        self.task_num = len(task_list)

        if isinstance(task_list, str) or self.task_num < 2:
            raise MLGBError
        for _task in task_list:
            if _task not in ('binary', 'regression'):
                raise MLGBError
        if pool_mv_mode not in MVPoolModeList:
            raise MLGBError('pool_mv_mode')
        if pool_seq_mode not in MVPoolModeList:
            raise MLGBError('pool_seq_mode')
        if not embed_dim:
            raise MLGBError('embed_dim')
        if share_mode not in ('SB:hard', 'SB:soft'):
            raise MLGBError

        self.input_fn = InputsLayer(
            feature_names=feature_names,
            inputs_mode='Inputs:feature',
            inputs_if_multivalued=inputs_if_multivalued,
            inputs_if_sequential=inputs_if_sequential,
            inputs_if_embed_dense=inputs_if_embed_dense,
            outputs_dense_if_add_sparse=True,
            embed_dim=embed_dim,
            embed_l2=embed_l2,
            embed_initializer=embed_initializer,
            embed_2d_dim=embed_2d_dim,
            embed_cate_if_output2d=False,
            pool_mv_mode=pool_mv_mode,
            pool_mv_axis=pool_mv_axis,
            pool_mv_l2=pool_mv_l2,
            pool_mv_initializer=pool_mv_initializer,
            pool_seq_mode=pool_seq_mode,
            pool_seq_axis=pool_seq_axis,
            pool_seq_l2=pool_seq_l2,
            pool_seq_initializer=pool_seq_initializer,
            seed=seed,
        )
        self.task_fn_list = [
            TaskLayer(
                task=task_list[i],
                task_linear_if_identity=True,
                seed=seed + i if isinstance(seed, int) else seed,
            )
            for i in range(self.task_num)
        ]
        self.sb_fn = SharedBottomLayer(
            task_fn_list=self.task_fn_list,
            share_mode=share_mode,
            hard_bottom_dnn_hidden_units=hard_bottom_dnn_hidden_units,
            hard_tower_dnn_hidden_units=hard_tower_dnn_hidden_units,
            soft_dnn_hidden_units=soft_dnn_hidden_units,
            dnn_activation=dnn_activation,
            dnn_dropout=dnn_dropout,
            dnn_if_bn=dnn_if_bn,
            dnn_if_ln=dnn_if_ln,
            dnn_l2=dnn_l2,
            dnn_initializer=dnn_initializer,
            seed=seed,
        )

    @tf.function
    def call(self, inputs):
        dense_2d_tensor, _ = self.input_fn(inputs)

        outputs = self.sb_fn(dense_2d_tensor)
        return outputs


class ESMM(tf.keras.Model):
    def __init__(self, feature_names, task=('binary', 'binary'), seed=None,
                 inputs_if_multivalued=False, inputs_if_sequential=False, inputs_if_embed_dense=False,
                 embed_dim=32, embed_2d_dim=None, embed_l2=0.0, embed_initializer=None,
                 pool_mv_mode='Pooling:average', pool_mv_axis=2, pool_mv_l2=0.0, pool_mv_initializer=None,
                 pool_seq_mode='Pooling:average', pool_seq_axis=1, pool_seq_l2=0.0, pool_seq_initializer=None,
                 dnn_hidden_units=(64, 32), dnn_activation='selu', dnn_dropout=0.0, dnn_if_bn=False, dnn_if_ln=False,
                 dnn_l2=0.0, dnn_initializer=None,
                 ):
        """
        Model Name: ESMM(EntireSpaceMultitaskModel)
        Paper Team: Alibaba
        Paper Year: 2018
        Paper Name: <Entire Space Multi-Task Model: An Effective Approach for Estimating Post-Click Conversion Rate>
        Paper Link: https://arxiv.org/pdf/1804.07931.pdf

        Task Inputs Parameters:
            :param feature_names: tuple(tuple(dict)), must. Embedding need vocabulary size and custom embed_dim of features.
            :param task: Tuple[str], default ('binary', 'binary').
            :param seed: Optional[int], default None.
            :param inputs_if_multivalued: bool, default False.
            :param inputs_if_sequential: bool, default False.
            :param inputs_if_embed_dense: bool, default False.
            :param embed_dim: int, default 32.
            :param embed_2d_dim: Optional[int], default None. When None, each field has own embed_dim by feature_names.
            :param embed_l2: float, default 0.0.
            :param embed_initializer: Optional[str], default None. When None, activation judge first, he_uniform end.
            :param pool_mv_mode: str, default 'Pooling:average'. Pooling mode of multivalued inputs. Union[
                                'Attention', 'Weighted', 'Pooling:max', 'Pooling:average', 'Pooling:sum']
            :param pool_mv_axis: int, default 2. Pooling axis of multivalued inputs.
            :param pool_mv_l2: float, default 0.0. When pool_mv_mode is in ('Weighted', 'Attention'), it works.
            :param pool_mv_initializer: Optional[str], default None. When None, activation judge first,
                                he_uniform end. When pool_mv_mode is in ('Weighted', 'Attention'), it works.
            :param pool_seq_mode: str, default 'Pooling:average'. Pooling mode of sequential inputs. Union[
                                'Attention', 'Weighted', 'Pooling:max', 'Pooling:average', 'Pooling:sum']
            :param pool_seq_axis: int, default 1. Pooling axis of sequential inputs.
            :param pool_seq_l2: float, default 0.0. When pool_seq_mode is in ('Weighted', 'Attention'), it works.
            :param pool_seq_initializer: Optional[str], default None. When None, activation judge first,
                                he_uniform end. When pool_seq_mode is in ('Weighted', 'Attention'), it works.

        Task Model Parameters:
            :param dnn_hidden_units: Tuple[int], default (64, 32).
            :param dnn_activation: Optional[str], default 'selu'.
            :param dnn_dropout: float, default 0.0.
            :param dnn_if_bn: bool, default False. Batch Normalization.
            :param dnn_if_ln: bool, default False. Layer Normalization.
            :param dnn_l2: float, default 0.0.
            :param dnn_initializer: Optional[str], default None. When None, activation judge first, he_uniform end.
        """
        super().__init__()
        task_list = task
        self.task_num = len(task_list)

        if isinstance(task_list, str) or self.task_num < 2:
            raise MLGBError
        for _task in task_list:
            if _task != 'binary':
                raise MLGBError
        if pool_mv_mode not in MVPoolModeList:
            raise MLGBError('pool_mv_mode')
        if pool_seq_mode not in MVPoolModeList:
            raise MLGBError('pool_seq_mode')
        if not embed_dim:
            raise MLGBError('embed_dim')

        self.input_fn = InputsLayer(
            feature_names=feature_names,
            inputs_mode='Inputs:feature',
            inputs_if_multivalued=inputs_if_multivalued,
            inputs_if_sequential=inputs_if_sequential,
            inputs_if_embed_dense=inputs_if_embed_dense,
            outputs_dense_if_add_sparse=True,
            embed_dim=embed_dim,
            embed_l2=embed_l2,
            embed_initializer=embed_initializer,
            embed_2d_dim=embed_2d_dim,
            embed_cate_if_output2d=False,
            pool_mv_mode=pool_mv_mode,
            pool_mv_axis=pool_mv_axis,
            pool_mv_l2=pool_mv_l2,
            pool_mv_initializer=pool_mv_initializer,
            pool_seq_mode=pool_seq_mode,
            pool_seq_axis=pool_seq_axis,
            pool_seq_l2=pool_seq_l2,
            pool_seq_initializer=pool_seq_initializer,
            seed=seed,
        )
        self.task_fn_list = [
            TaskLayer(
                task=task_list[i],
                task_linear_if_identity=True,
                seed=seed + i if isinstance(seed, int) else seed,
            )
            for i in range(self.task_num)
        ]
        self.esmm_fn = EntireSpaceMultitaskModelLayer(
            task_fn_list=self.task_fn_list,
            dnn_hidden_units=dnn_hidden_units,
            dnn_activation=dnn_activation,
            dnn_dropout=dnn_dropout,
            dnn_if_bn=dnn_if_bn,
            dnn_if_ln=dnn_if_ln,
            dnn_l2=dnn_l2,
            dnn_initializer=dnn_initializer,
            seed=seed,
        )

    @tf.function
    def call(self, inputs):
        dense_2d_tensor, _ = self.input_fn(inputs)

        outputs = self.esmm_fn(dense_2d_tensor)
        return outputs


class MMoE(tf.keras.Model):
    def __init__(self, feature_names, task=('binary', 'binary'), seed=None,
                 inputs_if_multivalued=False, inputs_if_sequential=False, inputs_if_embed_dense=False,
                 embed_dim=32, embed_2d_dim=None, embed_l2=0.0, embed_initializer=None,
                 pool_mv_mode='Pooling:average', pool_mv_axis=2, pool_mv_l2=0.0, pool_mv_initializer=None,
                 pool_seq_mode='Pooling:average', pool_seq_axis=1, pool_seq_l2=0.0, pool_seq_initializer=None,
                 expert_num=4,
                 expert_dnn_hidden_units=(64, 32), gate_dnn_hidden_units=(32,), tower_dnn_hidden_units=(32,),
                 dnn_activation='selu', dnn_dropout=0.0, dnn_if_bn=False, dnn_if_ln=False,
                 dnn_l2=0.0, dnn_initializer=None,
                 ):
        """
        Model Name: MMoE(MultigateMixtureOfExpert)
        Paper Team: Google(Alphabet)
        Paper Year: 2018
        Paper Name: <Modeling Task Relationships in Multi-task Learning with Multi-gate Mixture-of-Experts>
        Paper Link: https://dl.acm.org/doi/pdf/10.1145/3219819.3220007

        Task Inputs Parameters:
            :param feature_names: tuple(tuple(dict)), must. Embedding need vocabulary size and custom embed_dim of features.
            :param task: Tuple[str], default ('binary', 'binary').
            :param seed: Optional[int], default None.
            :param inputs_if_multivalued: bool, default False.
            :param inputs_if_sequential: bool, default False.
            :param inputs_if_embed_dense: bool, default False.
            :param embed_dim: int, default 32.
            :param embed_2d_dim: Optional[int], default None. When None, each field has own embed_dim by feature_names.
            :param embed_l2: float, default 0.0.
            :param embed_initializer: Optional[str], default None. When None, activation judge first, he_uniform end.
            :param pool_mv_mode: str, default 'Pooling:average'. Pooling mode of multivalued inputs. Union[
                                'Attention', 'Weighted', 'Pooling:max', 'Pooling:average', 'Pooling:sum']
            :param pool_mv_axis: int, default 2. Pooling axis of multivalued inputs.
            :param pool_mv_l2: float, default 0.0. When pool_mv_mode is in ('Weighted', 'Attention'), it works.
            :param pool_mv_initializer: Optional[str], default None. When None, activation judge first,
                                he_uniform end. When pool_mv_mode is in ('Weighted', 'Attention'), it works.
            :param pool_seq_mode: str, default 'Pooling:average'. Pooling mode of sequential inputs. Union[
                                'Attention', 'Weighted', 'Pooling:max', 'Pooling:average', 'Pooling:sum']
            :param pool_seq_axis: int, default 1. Pooling axis of sequential inputs.
            :param pool_seq_l2: float, default 0.0. When pool_seq_mode is in ('Weighted', 'Attention'), it works.
            :param pool_seq_initializer: Optional[str], default None. When None, activation judge first,
                                he_uniform end. When pool_seq_mode is in ('Weighted', 'Attention'), it works.

        Task Model Parameters:
            :param expert_num: int, default 4.
            :param expert_dnn_hidden_units: Tuple[int], default (64, 32).
            :param gate_dnn_hidden_units: Tuple[int], default (32,).
            :param tower_dnn_hidden_units: Tuple[int], default (32,).
            :param dnn_activation: Optional[str], default 'selu'.
            :param dnn_dropout: float, default 0.0.
            :param dnn_if_bn: bool, default False. Batch Normalization.
            :param dnn_if_ln: bool, default False. Layer Normalization.
            :param dnn_l2: float, default 0.0.
            :param dnn_initializer: Optional[str], default None. When None, activation judge first, he_uniform end.
        """
        super().__init__()
        task_list = task
        self.task_num = len(task_list)

        if isinstance(task_list, str) or self.task_num < 2:
            raise MLGBError
        for _task in task_list:
            if _task not in ('binary', 'regression'):
                raise MLGBError
        if pool_mv_mode not in MVPoolModeList:
            raise MLGBError('pool_mv_mode')
        if pool_seq_mode not in MVPoolModeList:
            raise MLGBError('pool_seq_mode')
        if not embed_dim:
            raise MLGBError('embed_dim')

        self.input_fn = InputsLayer(
            feature_names=feature_names,
            inputs_mode='Inputs:feature',
            inputs_if_multivalued=inputs_if_multivalued,
            inputs_if_sequential=inputs_if_sequential,
            inputs_if_embed_dense=inputs_if_embed_dense,
            outputs_dense_if_add_sparse=True,
            embed_dim=embed_dim,
            embed_l2=embed_l2,
            embed_initializer=embed_initializer,
            embed_2d_dim=embed_2d_dim,
            embed_cate_if_output2d=False,
            pool_mv_mode=pool_mv_mode,
            pool_mv_axis=pool_mv_axis,
            pool_mv_l2=pool_mv_l2,
            pool_mv_initializer=pool_mv_initializer,
            pool_seq_mode=pool_seq_mode,
            pool_seq_axis=pool_seq_axis,
            pool_seq_l2=pool_seq_l2,
            pool_seq_initializer=pool_seq_initializer,
            seed=seed,
        )
        self.task_fn_list = [
            TaskLayer(
                task=task_list[i],
                task_linear_if_identity=True,
                seed=seed + i if isinstance(seed, int) else seed,
            )
            for i in range(self.task_num)
        ]
        self.mmoe_fn = MultigateMixtureOfExpertLayer(
            task_fn_list=self.task_fn_list,
            expert_num=expert_num,
            expert_dnn_hidden_units=expert_dnn_hidden_units,
            gate_dnn_hidden_units=gate_dnn_hidden_units,
            tower_dnn_hidden_units=tower_dnn_hidden_units,
            dnn_activation=dnn_activation,  # dead relu.
            dnn_dropout=dnn_dropout,
            dnn_if_bn=dnn_if_bn,
            dnn_if_ln=dnn_if_ln,
            dnn_l2=dnn_l2,
            dnn_initializer=dnn_initializer,
            seed=seed,
        )

    @tf.function
    def call(self, inputs):
        dense_2d_tensor, _ = self.input_fn(inputs)

        outputs = self.mmoe_fn(dense_2d_tensor)
        return outputs


class PLE(tf.keras.Model):
    def __init__(self, feature_names, task=('binary', 'binary'), seed=None,
                 inputs_if_multivalued=False, inputs_if_sequential=False, inputs_if_embed_dense=False,
                 embed_dim=32, embed_2d_dim=None, embed_l2=0.0, embed_initializer=None,
                 pool_mv_mode='Pooling:average', pool_mv_axis=2, pool_mv_l2=0.0, pool_mv_initializer=None,
                 pool_seq_mode='Pooling:average', pool_seq_axis=1, pool_seq_l2=0.0, pool_seq_initializer=None,
                 tower_dnn_hidden_units=(64, 32),
                 expert_men_num=4, shared_men_num=4, men_hidden_units=(64, 32),
                 dnn_activation='selu', dnn_dropout=0.0, dnn_if_bn=False, dnn_if_ln=False,
                 dnn_l2=0.0, dnn_initializer=None,
                 ):
        """
        Model Name: PLE(ProgressiveLayeredExtraction)
        Paper Team: Tencent
        Paper Year: 2020
        Paper Name: <Progressive Layered Extraction (PLE): A Novel Multi-Task Learning (MTL) Model for Personalized Recommendations>
        Paper Link: https://www.sci-hub.se/10.1145/3383313.3412236

        Task Inputs Parameters:
            :param feature_names: tuple(tuple(dict)), must. Embedding need vocabulary size and custom embed_dim of features.
            :param task: Tuple[str], default ('binary', 'binary').
            :param seed: Optional[int], default None.
            :param inputs_if_multivalued: bool, default False.
            :param inputs_if_sequential: bool, default False.
            :param inputs_if_embed_dense: bool, default False.
            :param embed_dim: int, default 32.
            :param embed_2d_dim: Optional[int], default None. When None, each field has own embed_dim by feature_names.
            :param embed_l2: float, default 0.0.
            :param embed_initializer: Optional[str], default None. When None, activation judge first, he_uniform end.
            :param pool_mv_mode: str, default 'Pooling:average'. Pooling mode of multivalued inputs. Union[
                                'Attention', 'Weighted', 'Pooling:max', 'Pooling:average', 'Pooling:sum']
            :param pool_mv_axis: int, default 2. Pooling axis of multivalued inputs.
            :param pool_mv_l2: float, default 0.0. When pool_mv_mode is in ('Weighted', 'Attention'), it works.
            :param pool_mv_initializer: Optional[str], default None. When None, activation judge first,
                                he_uniform end. When pool_mv_mode is in ('Weighted', 'Attention'), it works.
            :param pool_seq_mode: str, default 'Pooling:average'. Pooling mode of sequential inputs. Union[
                                'Attention', 'Weighted', 'Pooling:max', 'Pooling:average', 'Pooling:sum']
            :param pool_seq_axis: int, default 1. Pooling axis of sequential inputs.
            :param pool_seq_l2: float, default 0.0. When pool_seq_mode is in ('Weighted', 'Attention'), it works.
            :param pool_seq_initializer: Optional[str], default None. When None, activation judge first,
                                he_uniform end. When pool_seq_mode is in ('Weighted', 'Attention'), it works.

        Task Model Parameters:
            :param tower_dnn_hidden_units: Tuple[int], default (64, 32).
            :param expert_men_num: int, default 4.
            :param shared_men_num: int, default 4.
            :param men_hidden_units: Tuple[int], default (64, 32).
            :param dnn_activation: Optional[str], default 'selu'.
            :param dnn_dropout: float, default 0.0.
            :param dnn_if_bn: bool, default False. Batch Normalization.
            :param dnn_if_ln: bool, default False. Layer Normalization.
            :param dnn_l2: float, default 0.0.
            :param dnn_initializer: Optional[str], default None. When None, activation judge first, he_uniform end.
        """
        super().__init__()
        task_list = task
        self.task_num = len(task_list)

        if isinstance(task_list, str) or self.task_num < 2:
            raise MLGBError
        for _task in task_list:
            if _task not in ('binary', 'regression'):
                raise MLGBError
        if pool_mv_mode not in MVPoolModeList:
            raise MLGBError('pool_mv_mode')
        if pool_seq_mode not in MVPoolModeList:
            raise MLGBError('pool_seq_mode')
        if not embed_dim:
            raise MLGBError('embed_dim')

        self.input_fn = InputsLayer(
            feature_names=feature_names,
            inputs_mode='Inputs:feature',
            inputs_if_multivalued=inputs_if_multivalued,
            inputs_if_sequential=inputs_if_sequential,
            inputs_if_embed_dense=inputs_if_embed_dense,
            outputs_dense_if_add_sparse=True,
            embed_dim=embed_dim,
            embed_l2=embed_l2,
            embed_initializer=embed_initializer,
            embed_2d_dim=embed_2d_dim,
            embed_cate_if_output2d=False,
            pool_mv_mode=pool_mv_mode,
            pool_mv_axis=pool_mv_axis,
            pool_mv_l2=pool_mv_l2,
            pool_mv_initializer=pool_mv_initializer,
            pool_seq_mode=pool_seq_mode,
            pool_seq_axis=pool_seq_axis,
            pool_seq_l2=pool_seq_l2,
            pool_seq_initializer=pool_seq_initializer,
            seed=seed,
        )
        self.task_fn_list = [
            TaskLayer(
                task=task_list[i],
                task_linear_if_identity=True,
                seed=seed + i if isinstance(seed, int) else seed,
            )
            for i in range(self.task_num)
        ]
        self.ple_fn = ProgressiveLayeredExtractionLayer(
            task_fn_list=self.task_fn_list,
            tower_dnn_hidden_units=tower_dnn_hidden_units,
            expert_men_num=expert_men_num,
            shared_men_num=shared_men_num,
            men_hidden_units=men_hidden_units,
            dnn_activation=dnn_activation,
            dnn_dropout=dnn_dropout,
            dnn_if_bn=dnn_if_bn,
            dnn_if_ln=dnn_if_ln,
            dnn_l2=dnn_l2,
            dnn_initializer=dnn_initializer,
            seed=seed,
        )

    @tf.function
    def call(self, inputs):
        dense_2d_tensor, _ = self.input_fn(inputs)

        outputs = self.ple_fn(dense_2d_tensor)
        return outputs


class PEPNet(tf.keras.Model):
    def __init__(self, feature_names, task=('binary', 'binary'), seed=None,
                 inputs_if_multivalued=False, inputs_if_sequential=False, inputs_if_embed_dense=True,
                 embed_dim=32, embed_2d_dim=None, embed_l2=0.0, embed_initializer=None,
                 pool_mv_mode='Pooling:average', pool_mv_axis=2, pool_mv_l2=0.0, pool_mv_initializer=None,
                 pool_seq_mode='Pooling:average', pool_seq_axis=1, pool_seq_l2=0.0, pool_seq_initializer=None,
                 f_ep_id_list=(0,), f_pp_id_list=(1, 2, 3),
                 domain_num=1, tower_dnn_hidden_units=(64, 32),
                 gnu_factor=2.0, gnu_last_activation='sigmoid', gnu_if_concat_general_inputs=False,
                 dnn_activation='selu', dnn_dropout=0.0, dnn_if_bn=False, dnn_if_ln=False,
                 dnn_l2=0.0, dnn_initializer=None,
                 ):
        """
        Model Name: PEPNet(Parameter&EmbeddingPersonalizedNetwork)
        Paper Team: Kuaishou
        Paper Year: 2023
        Paper Name: <PEPNet: Parameter and Embedding Personalized Network for Infusing with Personalized Prior Information>
        Paper Link: https://arxiv.org/pdf/2302.01115.pdf

        Task Inputs Parameters:
            :param feature_names: tuple(tuple(dict)), must. Embedding need vocabulary size and custom embed_dim of features.
            :param task: Tuple[str], default ('binary', 'binary').
            :param seed: Optional[int], default None.
            :param inputs_if_multivalued: bool, default False.
            :param inputs_if_sequential: bool, default False.
            :param inputs_if_embed_dense: bool, default True. Must True.
            :param embed_dim: int, default 32.
            :param embed_2d_dim: Optional[int], default None. When None, each field has own embed_dim by feature_names.
            :param embed_l2: float, default 0.0.
            :param embed_initializer: Optional[str], default None. When None, activation judge first, he_uniform end.
            :param pool_mv_mode: str, default 'Pooling:average'. Pooling mode of multivalued inputs. Union[
                                'Attention', 'Weighted', 'Pooling:max', 'Pooling:average', 'Pooling:sum']
            :param pool_mv_axis: int, default 2. Pooling axis of multivalued inputs.
            :param pool_mv_l2: float, default 0.0. When pool_mv_mode is in ('Weighted', 'Attention'), it works.
            :param pool_mv_initializer: Optional[str], default None. When None, activation judge first,
                                he_uniform end. When pool_mv_mode is in ('Weighted', 'Attention'), it works.
            :param pool_seq_mode: str, default 'Pooling:average'. Pooling mode of sequential inputs. Union[
                                'Attention', 'Weighted', 'Pooling:max', 'Pooling:average', 'Pooling:sum']
            :param pool_seq_axis: int, default 1. Pooling axis of sequential inputs.
            :param pool_seq_l2: float, default 0.0. When pool_seq_mode is in ('Weighted', 'Attention'), it works.
            :param pool_seq_initializer: Optional[str], default None. When None, activation judge first,
                                he_uniform end. When pool_seq_mode is in ('Weighted', 'Attention'), it works.

        Task Model Parameters:
            :param f_ep_id_list: Tuple[str], default (0,). Feature Index. About domain.
            :param f_pp_id_list: Tuple[str], default (1, 2, 3). Feature Index. About user, item, author.
            :param domain_num: int, default 1.
            :param tower_dnn_hidden_units: Tuple[int], default (64, 32).
            :param gnu_factor: float, default 2.0.
            :param gnu_last_activation: Optional[str], default 'sigmoid'.
            :param gnu_if_concat_general_inputs: bool, default False.
            :param dnn_activation: Optional[str], default 'selu'.
            :param dnn_dropout: float, default 0.0.
            :param dnn_if_bn: bool, default False. Batch Normalization.
            :param dnn_if_ln: bool, default False. Layer Normalization.
            :param dnn_l2: float, default 0.0.
            :param dnn_initializer: Optional[str], default None. When None, activation judge first, he_uniform end.
        """
        super().__init__()
        task_list = task
        self.task_num = len(task_list)

        if isinstance(task_list, str) or self.task_num < 2:
            raise MLGBError
        for _task in task_list:
            if _task not in ('binary', 'regression'):
                raise MLGBError
        if pool_mv_mode not in MVPoolModeList:
            raise MLGBError('pool_mv_mode')
        if pool_seq_mode not in MVPoolModeList:
            raise MLGBError('pool_seq_mode')
        if not embed_dim:
            raise MLGBError('embed_dim')
        if not inputs_if_embed_dense:
            raise MLGBError('inputs_if_embed_dense')
        if not domain_num > 0:
            raise MLGBError('domain_num')
        if not gnu_factor > 0:
            raise MLGBError('gnu_factor')
        if not (min(f_ep_id_list) >= 0 and min(f_pp_id_list) >= 0):
            raise MLGBError('f_ep_id_list or f_pp_id_list')

        self.f_ep_id_list = list(f_ep_id_list)
        self.f_pp_id_list = list(f_pp_id_list)

        self.input_fn = InputsLayer(
            feature_names=feature_names,
            inputs_mode='Inputs:feature',
            inputs_if_multivalued=inputs_if_multivalued,
            inputs_if_sequential=inputs_if_sequential,
            inputs_if_embed_dense=inputs_if_embed_dense,  #
            outputs_dense_if_add_sparse=True,
            embed_dim=embed_dim,
            embed_l2=embed_l2,
            embed_initializer=embed_initializer,
            embed_2d_dim=embed_2d_dim,
            embed_cate_if_output2d=False,
            pool_mv_mode=pool_mv_mode,
            pool_mv_axis=pool_mv_axis,
            pool_mv_l2=pool_mv_l2,
            pool_mv_initializer=pool_mv_initializer,
            pool_seq_mode=pool_seq_mode,
            pool_seq_axis=pool_seq_axis,
            pool_seq_l2=pool_seq_l2,
            pool_seq_initializer=pool_seq_initializer,
            seed=seed,
        )
        if domain_num > 1:
            self.task_fn_list = [
                [
                    TaskLayer(
                        task=task_list[j],
                        task_linear_if_identity=True,
                        seed=seed + i + j if isinstance(seed, int) else seed,
                    )
                    for i in range(domain_num)
                ]
                for j in range(self.task_num)
            ]
        else:
            self.task_fn_list = [
                TaskLayer(
                    task=task_list[i],
                    task_linear_if_identity=True,
                    seed=seed + i if isinstance(seed, int) else seed,
                )
                for i in range(self.task_num)
            ]
        self.pep_net_fn = ParameterEmbeddingPersonalizedNetworkLayer(
            task_fn_list=self.task_fn_list,
            domain_num=domain_num,
            tower_dnn_hidden_units=tower_dnn_hidden_units,
            gnu_factor=gnu_factor,
            gnu_last_activation=gnu_last_activation,
            gnu_if_concat_general_inputs=gnu_if_concat_general_inputs,
            dnn_activation=dnn_activation,
            dnn_dropout=dnn_dropout,
            dnn_if_bn=dnn_if_bn,
            dnn_if_ln=dnn_if_ln,
            dnn_l2=dnn_l2,
            dnn_initializer=dnn_initializer,
            seed=seed,
        )

    @tf.function
    def call(self, inputs):
        _, embed_3d_tensor = self.input_fn(inputs)

        f_g_list = [i for i in range(embed_3d_tensor.shape[1]) if i not in set(self.f_ep_id_list + self.f_pp_id_list)]
        g_3d_tensor = tf.gather(embed_3d_tensor, indices=tf.constant(f_g_list, dtype=tf.int32), axis=1)
        ep_3d_tensor = tf.gather(embed_3d_tensor, indices=tf.constant(self.f_ep_id_list, dtype=tf.int32), axis=1)
        pp_3d_tensor = tf.gather(embed_3d_tensor, indices=tf.constant(self.f_pp_id_list, dtype=tf.int32), axis=1)

        outputs = self.pep_net_fn([g_3d_tensor, ep_3d_tensor, pp_3d_tensor])
        return outputs






























# coding=utf-8
# author=uliontse

import tensorflow as tf
if tf.__version__ >= '2.16':
    del tf
    import os
    import sys
    os.environ["TF_USE_LEGACY_KERAS"] = '1'
    sys.stderr.write('Using Keras2 backend.\n')  # `pip install tf-keras~=2.16`

import numpy
import pandas
import tensorflow as tf
from sklearn.metrics import confusion_matrix, classification_report

from mlgb import get_model
from mlgb.data import get_multitask_label_data
from mlgb.utils import check_filepath


if __name__ == '__main__':
    model_name = 'PEPNet'
    seed = 0
    print(f'model_name: {model_name}')

    tmp_dir = '.tmp'
    model_dir = f'{tmp_dir}/{model_name}_tf'
    log_dir = f'{model_dir}/log_dir'
    save_model_dir = f'{model_dir}/save_model'
    check_filepath(tmp_dir, model_dir, log_dir, save_model_dir)

    device = 'cuda' if tf.test.is_built_with_cuda() else 'cpu'
    print(f'device: {device}')

    feature_names, (x_train, y_train), (x_test, y_test) = get_multitask_label_data(
        n_samples=int(1e4),
        negative_class_weight=0.9,
        multitask_cvr=0.5,
        test_size=0.15,
        seed=seed,
    )
    y_train = (y_train[:, 0], y_train[:, 1])
    y_test = (y_test[:, 0], y_test[:, 1])
    print([len(names) for names in feature_names])

    model = get_model(
        feature_names=feature_names,
        model_name=model_name,
        task=('binary', 'binary'),
        aim='mtl',
        lang='tf',
        seed=seed,
    )
    model.compile(
        loss=[tf.losses.BinaryCrossentropy(), tf.losses.BinaryCrossentropy()],
        optimizer=tf.optimizers.Nadam(learning_rate=1e-3),
        metrics=[
            # tfa.metrics.F1Score(num_classes=1, threshold=0.5, average='macro'),  # import tensorflow_addons as tfa
            # tf.metrics.F1Score(threshold=0.5, average='macro'),  # tf.__version__ >= '2.13'
            tf.metrics.AUC(),
        ],
    )
    history = model.fit(
        x=x_train,
        y=y_train,
        batch_size=32,
        epochs=10,
        validation_split=0.15,
        class_weight=None,
        callbacks=[
            tf.keras.callbacks.TensorBoard(log_dir=log_dir),
            tf.keras.callbacks.EarlyStopping(monitor='val_loss', patience=10, min_delta=1e-3),
        ]
    )

    print(model.summary())
    print(pandas.DataFrame(history.history))

    test_evaluate = model.evaluate(x=x_test, y=y_test, batch_size=None, return_dict=True)
    print(test_evaluate)

    y_pred_prob = model.predict(x_test)
    y_pred_prob = numpy.squeeze(y_pred_prob)  # flatten
    y_pred = numpy.where(y_pred_prob >= 0.5, 1, 0)
    print(y_pred.shape)

    print('CTR:')
    print(confusion_matrix(y_test[0], y_pred[0]))
    print(classification_report(y_test[0], y_pred[0]))

    print('CVR:')
    print(confusion_matrix(y_test[1], y_pred[1]))
    print(classification_report(y_test[1], y_pred[1]))

    # ranking model:
    model.save(filepath=save_model_dir, save_format='tf')
    local_model = tf.keras.models.load_model(filepath=save_model_dir)

    y_local_pred = local_model.predict(x_test)
    y_model_pred = model.predict(x_test)

    print('y_local_ctr_pred == y_model_ctr_pred:', numpy.allclose(y_local_pred[0], y_model_pred[0]))
    print(numpy.abs(numpy.array(y_local_pred[0]) - numpy.array(y_model_pred[0])).max())

    print('y_local_cvr_pred == y_model_cvr_pred:', numpy.allclose(y_local_pred[1], y_model_pred[1]))
    print(numpy.abs(numpy.array(y_local_pred[1]) - numpy.array(y_model_pred[1])).max())
